/* DChub - a Direct Connect Hub for Linux
 * Copyright (C) 2001 Eric Prevoteau
 *
 * main.c: Copyright (C) Eric Prevoteau <www@ac2i.tzo.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>
#include <signal.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#ifdef HAVE_GETOPT_LONG
#include <getopt.h>
#endif
#include <errno.h>
#include <glib.h>

#include "config.h"
#include "main.h"
#include "network.h"
#include "manage_cnx.h"
#include "do_hub_reg.h"

#include "cnx_lst.h"
#include "emb_perl.h"
#ifdef WITH_PYTHON
#include "emb_python.h"
#endif
#include "hub_to_hub.h"
#include "timed_out_string.h"
#include "tos_key.h"
#include "hub_cluster.h"
#include <pwd.h>
#include <passwd.h>
#include "color_debug.h"
#include <regex.h>
#include "gvar.h"

static int main_sck;
time_t gl_cur_time;

unsigned short hub_port=411;
unsigned short reg_hub_port=0;

GString *hubname;
char *tos_filename=NULL;

int	gl_sr_interval;
int	gl_moderate_mode;
int	gl_silent_limit;
int	gl_kick_limit;
char	*gl_bind_addr;
regex_t *gl_lock_patern = NULL;
int	is_ip(char *str);

/***************************/
/* configuration file vars */
/***************************/
char *conf_filename=NULL;		/* name of the configuration file (to use with edb) */
t_db_sym **conf_file=NULL;
G_LOCK_DEFINE(conf_file);

/******************/
/* user file vars */
/******************/
char *user_filename=NULL;		/* name of the user file */
char *user_uid=NULL;		/* name of the user file */
FILE *user_file=NULL;
G_LOCK_DEFINE(user_file);

/***************************************/
/* some vars used for Perl interpreter */
/***************************************/
char *perl_script_dir=NULL;
char *perl_script_main=NULL;

#ifdef WITH_PYTHON
/*****************************************/
/* some vars used for Python interpreter */
/*****************************************/
char *python_script_dir  = NULL;
char *python_main_script = NULL;
#endif

/*****************************************/
/* somes vars used for external programs */
/*****************************************/
char *ext_prog_dir=NULL;

/****************************************************************************/
/* it is not possible to start an external program from user thread because */
/* starting a program adds a user and this requires write access on cnx_lst */
/* Unfortunately, when a user thread manages its connection, it sets a read */
/* lock on the structure.                                                   */
/****************************************************************************/
GPtrArray *prog_to_start=NULL;

/*******************************************************/
/* add an external program to start by the main thread */
/*******************************************************/
void add_prog_to_start(char *progname)
{
	if(prog_to_start==NULL)
		prog_to_start=g_ptr_array_new();

	g_ptr_array_add(prog_to_start,strdup(progname));
}

/**************************************************************************/
/* start a new external program and attach it to the hub as a normal user */
/******************************************************************************/
/* the external program is always seen as an operator (not master of the hub) */
/******************************************************************************/
/* progname: name of the program (contains only valid DC char) */
/***************************************************************/
static void start_new_ext_prog(char *progname)
{
	int sockfd[2];			/* 0=hub side, 1=program side */
	CNX_ENTRY *ptr;
	int num;

	if(ext_prog_dir==NULL)
	{
		fprintf(stderr,"No external program directory defined.\n");
		return;
	}

	/* create thread entry */
	ptr=malloc(sizeof(CNX_ENTRY));
	if(ptr==NULL)
	{
		fprintf(stderr,"Out of memory, malloc fail.\n");
		return;
	}
	memset(ptr, 0, sizeof(CNX_ENTRY));

	/* allocate a socketpair (one side for the hub, the other for the program) */
	if(socketpair(AF_INET,SOCK_STREAM,0,sockfd)==-1)
	{
#ifndef EOPNOSUPPORT
#define EOPNOSUPPORT EOPNOTSUPP
#endif
		if(errno!=EOPNOSUPPORT)				/* according to the man, it should be this */
		{											/* but the value is not defined in errno.h */
			perror("start_new_ext_prog:socketpair");
			return;
		}
		else
		{
			int temp_sock;
			struct sockaddr_in lcl;
			int len_lcl=sizeof(lcl);

			perror("start_new_ext_prog:socketpair (trying manually)");

			/* linux doesn't seems to allow socketpair in INET domain, we will do this manually */
			/* but it is a bit long */

			/* create 2 sockets */
			sockfd[0]=_x_tcp(0);
			if(sockfd[0]==-1)
			{
				fail:
				return;
			}

			temp_sock=_x_tcp(0);
			if(temp_sock==-1)
			{
				close(sockfd[0]);
				goto fail;
			}

			if(getsockname(temp_sock,(void*)&lcl,&len_lcl)!=0)
			{
				perror("start_new_ext_prog:getsockname");
				close(temp_sock);
				close(sockfd[0]);
				goto fail;
			}
			listen(temp_sock,1);
			
			/* and connect the first on the second. The accept returns the true second */
			/* don't make a blocking connect else everything hangs */
			set_non_bloquant_sock(sockfd[0]);

			if(connect(sockfd[0],(void*)&lcl,len_lcl)!=0)
			{
				if(errno!=EINPROGRESS)
				{
					perror("start_new_ext_prog:connect");
					close(temp_sock);
					close(sockfd[0]);
					goto fail;
				}
			}
			
			sockfd[1]=accept(temp_sock,NULL,NULL);
			if(sockfd[1]==-1)
			{
				perror("start_new_ext_prog:accept");
				close(temp_sock);
				close(sockfd[0]);
				goto fail;
			}

			set_bloquant_sock(sockfd[0]);

			shutdown(temp_sock,2);
			close(temp_sock);

			/* that's all folks :) easy isn't it ? */
		}
	}

	ptr->sock_fd=sockfd[0];
	ptr->start_time=time(NULL);
	ptr->handshake_done=0;

	/* user information */
	ptr->user_cnx_type="Bot";
	ptr->user_nick=g_string_new(progname);
	ptr->shared_size=0;
	ptr->user_mail=g_string_new("");
	ptr->user_description=g_string_new("");
	ptr->user_flag=1;
	ptr->is_op= OP_PRIV | BOT_PRIV;
	ptr->client_version=g_string_new("");
	ptr->incoming_commands=NULL;
	ptr->in_partial=NULL;
	ptr->outgoing_commands=NULL;

	ptr->ip.s_addr = 0;
	ptr->server_id = 0; /* this mean a local connected user */
	ptr->connected = 0;
	ptr->next = 0;
	ptr->created_lock = 0;
	ptr->temp_nick = 0;
	ptr->user_pass = 0;
	ptr->log_level = 5;
	ptr->tmp_op_type = 0;
	add_new_cnx_entry(ptr,1);

	switch( num=fork())
	{
		case -1:
					/* bad news */
					printf("fork fails: %s\n",progname);
					shutdown(sockfd[1],2);
					close(sockfd[1]);
					/* let the main loop handles structure deletion */
					break;

		case 0:	/* the son */
					{
						GPtrArray *arg_exec;
						char buf[64];
						GString *str_hub_ip;

						/* close all unnecessary descriptor */
						{
							int i;
							int mx;

							mx=sysconf(_SC_OPEN_MAX);
							for(i=3;i<mx;i++)
								if(i!=sockfd[1])
									close(i);
						}

						sprintf(buf,"%u",sockfd[1]);

						str_hub_ip=get_default_host_ip(gl_bind_addr);
						/* create the arguments of the program to start */
						arg_exec=g_ptr_array_new();
						g_ptr_array_add(arg_exec,g_strconcat(ext_prog_dir,"/",progname,NULL));
						g_ptr_array_add(arg_exec,progname);
						g_ptr_array_add(arg_exec,buf);
						g_ptr_array_add(arg_exec,str_hub_ip->str);
						g_ptr_array_add(arg_exec,NULL);
						
						execv(g_ptr_array_index(arg_exec,0),(void*)(arg_exec->pdata));
						perror("execv");
						printf("execv fails: %s %s %s\n",(char*)g_ptr_array_index(arg_exec,0),progname,buf);
					}
					/* bad news */
					shutdown(sockfd[1],2);
					close(sockfd[1]);
					/* let the main loop handles structure deletion */
					_exit(1);
					break;

		default:
					{	/* notify to all users the incoming of this program */
						GString *msg;
						GString_ref *msg2;
						int i;
#ifdef __CYGWIN__
						char		long_int[48];
#endif


						msg=g_string_new("");
						/* send the hello string to everyone */
						g_string_sprintf(msg,"$Hello %s|",ptr->user_nick->str);
						msg2 = conv_gstring_to_gstring_ref(msg);
						queue_to_everyone(msg2);
						g_string_ref_free(msg2);


						msg=g_string_new("");
						/* and also the myinfo string */
#ifdef __CYGWIN__
						sprintf(long_int, "%Lu", ptr->shared_size);
						g_string_sprintf(msg,"$MyINFO $ALL %s %s$ $%s%c$%s$%s$|",
								 ptr->user_nick->str,
								 ptr->user_description->str,
								 ptr->user_cnx_type,
								 (char)ptr->user_flag,
								 ptr->user_mail->str,
								 long_int);
						
#else
						g_string_sprintf(msg,"$MyINFO $ALL %s %s$ $%s%c$%s$%Lu$|",
									ptr->user_nick->str,
									ptr->user_description->str,
									ptr->user_cnx_type,
									(char)ptr->user_flag,
									ptr->user_mail->str,
									ptr->shared_size);
#endif
						msg2 = conv_gstring_to_gstring_ref(msg);
						queue_to_everyone(msg2);
						g_string_ref_free(msg2);

						//queue_to_extern_everyone(msg->str,ptr->user_nick->str); FIXED ?

						/* and to conclude, update oplist of all users */
						msg=g_string_new("$OpList ");

						for(i=0;i<HASH_SIZE;i++)
						{
							CNX_ENTRY *ent;
							for(ent=cnx_lst->entry[i];ent;ent=ent->next)
							{
								if(ent->is_op && OP_PRIV)
								{
									msg=g_string_append(msg,ent->user_nick->str);
									msg=g_string_append(msg,"$$");
								}
							}
						}
						msg=g_string_append_c(msg,'|');
						msg2 = conv_gstring_to_gstring_ref(msg);
						queue_to_everyone(msg2);
						g_string_ref_free(msg2);
					}
					close(sockfd[1]);			/* close the remote side */
					printf("extproc ID: %d\n",num);
					break;
	}
	/* after the fork, we assume that handshake is done */
	ptr->handshake_done=1;
	inject_wait_table(cnx_lst);
}

/**************************************************************/
/* check periodically if any external program must be started */
/**************************************************************/
static void check_prog_to_start(void)
{
	if(prog_to_start==NULL)
		return;

	while(prog_to_start->len!=0)
	{
		char *progname;

		progname=g_ptr_array_remove_index_fast(prog_to_start,0);
		if(progname!=NULL)
		{
			start_new_ext_prog(progname);
			free(progname);
		}
	}
}

/******************************/
/* create a valid $Lock value */
/*******************************/
char *create_a_lock(void)
{
	int i;
	int ln;
	char *lck;

	lck = malloc(100);
	ln=50+rand()%50;

	for(i=0;i<ln;i++)
	{
		lck[i]='%'+rand()%('z'-'%');	/* create a value between '%' and 'z' */
	}
	lck[ln]='\0';
	return (lck);
}


static void	new_incoming_connexion(void)
{
	int sock_fd;
	CNX_ENTRY *ptr;
	struct sockaddr_in remote;
	int x=sizeof(remote);
	char *dummy1;
	int dummy_len;

	sock_fd=accept(main_sck,(struct sockaddr *)&remote,&x);
	if(sock_fd==-1)
		return;

	/* we first check if someone doesn't try to flood the hub with dummy connection */

	if(get_tos_entry(CNX_TOSKEY,(char *)&(remote.sin_addr),sizeof(remote.sin_addr),0,&dummy1,&dummy_len)==1)
	{
		/*printf(stderr,"Connection attempt from the same ip in less than 10 seconds.\n"); */
		shutdown(sock_fd,2);
		close(sock_fd);
		return;
	}

	if(get_tos_entry(TBAN_TOSKEY,(char *)&(remote.sin_addr),sizeof(remote.sin_addr),0,&dummy1,&dummy_len)==1)
	{
		/*	fprintf(stderr,"Connection attempt from a temporary banned ip.\n"); */
		shutdown(sock_fd,2);
		close(sock_fd);
		return;
	}

	if(get_tos_entry(PBAN_TOSKEY,(char *)&(remote.sin_addr),sizeof(remote.sin_addr),0,&dummy1,&dummy_len)==1)
	{
		/* fprintf(stderr,"Connection attempt from a permanently banned ip.\n"); */
		shutdown(sock_fd,2);
		close(sock_fd);
		return;
	}

	/* put the address of the remote host in the CNX tos list */
	add_tos_entry(CNX_TOSKEY,10,(char *)&(remote.sin_addr),sizeof(remote.sin_addr),NULL,0);

	/* create thread entry */
	ptr=malloc(sizeof(CNX_ENTRY));
	if(ptr==NULL)
	{
		fprintf(stderr,"Out of memory, malloc fail.\n");
		shutdown(sock_fd,2);
		close(sock_fd);
		return;
	}
	memset(ptr, 0, sizeof(CNX_ENTRY));
	ptr->sock_fd=sock_fd;
	ptr->start_time=time(NULL);
	ptr->handshake_done=0;

	/* user information */
	ptr->user_cnx_type="";
	ptr->user_nick=g_string_new("");
	ptr->shared_size=0;
	ptr->user_mail=g_string_new("");
	ptr->user_description=g_string_new("");
	ptr->user_flag=1;
	ptr->is_op=0;
	ptr->client_version=g_string_new("");

	ptr->incoming_commands=NULL;
	ptr->in_partial=NULL;
	ptr->outgoing_commands=NULL;

	ptr->ip = remote.sin_addr;
	ptr->server_id = 0; /* this mean a local connected user */
	ptr->connected = 0;
	ptr->next=NULL;
	ptr->log_level = 0;

	ptr->temp_nick = 0;
	ptr->user_pass = 0;
	ptr->log_level = 0;
	ptr->tmp_op_type = 0;


	ptr->created_lock = create_a_lock();
	{
		GString *tmp_buf=g_string_new("");
		GString_ref *tmp_buf2;

		g_string_sprintf(tmp_buf,"$Lock %s |",ptr->created_lock);
		tmp_buf2 = conv_gstring_to_gstring_ref(tmp_buf);
		append_new_cmd_to_glist_outgoing(ptr,&(ptr->outgoing_commands),tmp_buf2);
		g_string_ref_free(tmp_buf2);
	}

	add_new_cnx_entry(ptr,1);
}

static int user_wants_to_quit=0;

static void catch_sig(int sig)
{
	switch(sig)
	{
		case SIGTERM:
		case SIGQUIT:
		case SIGINT:
						user_wants_to_quit=1;
						break;
	}
}

/************************************/
/* we don't want to receive SIGPIPE */
/************************************/
static void set_sig(void)
{
	struct sigaction act;
	sigset_t set;

	/* ignore SIGPIPE */
	sigemptyset(&set);
	sigaddset(&set,SIGCHLD);
	sigaddset(&set,SIGPIPE);
	sigaddset(&set,SIGHUP);
	act.sa_handler=SIG_IGN;
	act.sa_mask=set;
	act.sa_flags=SA_RESTART;

	sigaction(SIGCHLD,&act,NULL);
	sigaction(SIGPIPE,&act,NULL);
	sigaction(SIGHUP,&act,NULL);
	sigprocmask(SIG_UNBLOCK,&set,NULL);

	/* and catch some other sig */
	act.sa_handler=catch_sig;
	act.sa_mask=set;
	act.sa_flags=SA_RESTART;
	sigaction(SIGINT,&act,NULL);
	sigaction(SIGQUIT,&act,NULL);
	sigaction(SIGTERM,&act,NULL);
	sigprocmask(SIG_UNBLOCK,&set,NULL);
}

static void	display_usage(char *prog)
{
	fprintf(stderr,"Usage: %s [options]\n"
						"Options:\n"
						" -h, --help                      Display this help\n"
						" -c, --conf=FILENAME             use the given filename as configuration file\n"
						"                                 (required)\n"
						" -u, --user=FILENAME             use the given filename as user/passwd file\n"
						"                                 (required)\n"
						" -U, --UID=USERNAME              ask DcHub to change UID after having bind the main port.\n"
						" -i, --init                      create the default configuration and user file\n"
						"                                 content. You must use this option the first\n"
						"                                 time you start the hub. Note: This function\n"
						"                                 deletes existing configuration and user file.\n"
						" -d, --pscriptdir=DIRECTORY      name of the directory containing perl script.\n"
						" -s, --pscriptinit=FILENAME      name of the filename loaded when perl starts.\n"
						" -n, --newport=NUMBER            change listening port of the hub. After using\n"
						"                                 this option, the new port is stored in the\n"
						"                                 database and the option is no more required.\n"
						" -f, --forceport=NUMBER          force DCHUB to register a specific Port\n"
						" -e, --exprogdir=DIRECTORY       name of the directory containing external programs.\n"
						"                                 (NOTE: if a file named 'AUTOSTART' exists in this\n"
						"                                 directory, this file contains the name of all\n"
						"                                 external programs to start (1 per line)\n"
						" -b, --bind=IP                   if the machin have more than one external IP\n"
#ifdef WITH_PYTHON
						" -p, --pyscriptdir=DIRECTORY     name of the directory containing Python scripts.\n"
						" -x, --pyinit=FILENAME           name of the filename loaded when Python starts.\n"
#endif
#ifndef HAVE_GETOPT_LONG
						"\n  WARNING this build do not support long arguments\n"
#endif
						"\n",prog);
}

static void parse_command_line_options(int argc,char **argv)
{
#ifdef HAVE_GETOPT_LONG
	static struct option optab[]= {
												{"help",no_argument,NULL,'h'},           /* get help */
												{"conf",required_argument,NULL,'c'},	/* define the configuration file */
												{"user",required_argument,NULL,'u'},	/* define the user/passwd file */
												{"UID",required_argument,NULL,'U'},	/* change UID after running */
												{"init",no_argument,NULL,'i'},			/* initialize configuration & user/passwd files to their default value */
												{"missing",no_argument,NULL,'m'},		/* add missing variables to the configuration & user/passwd files */
												{"pscriptdir",required_argument,NULL,'d'},	/* the Perl script dir */
												{"pscriptinit",required_argument,NULL,'s'},	/* the Perl script containing the main */
												{"newport",required_argument,NULL,'n'},	/* the new hub port to use */
												{"forceport",required_argument,NULL,'f'},	/* the new hub port to reg */
												{"exprogdir",required_argument,NULL,'e'},	/* the dir containing external program */
												{"bind",required_argument,NULL,'b'},	/*  */ 
#ifdef WITH_PYTHON
												{"pyscriptdir",required_argument,NULL,'p'},
												{"pyinit",required_argument,NULL,'x'},
#endif
												{NULL,0,NULL,'\0'}               /* last option */
											};
#endif
#ifdef WITH_PYTHON
	static const char *short_opt="hc:u:imd:s:n:e:U:f:p:xb:";
#else
	static const char *short_opt="hc:u:imd:s:n:e:U:f:b:";
#endif
	int do_init=0;
	int add_missing=0;
	int ch;
	struct stat st;
	int new_hub_port=-1;
	int tmp_reg_hub_port=-1;

	if(argc==1)
	{
		display_usage(argv[0]);
		exit(0);
	}
	opterr = 0;
#ifdef HAVE_GETOPT_LONG
	while((ch=getopt_long(argc,argv,short_opt,optab,NULL))!=EOF)
#else
	while((ch=getopt(argc,argv,short_opt))!=EOF)
#endif
	{
		switch(ch)
		{
			case 'h':
							display_usage(argv[0]);
							exit(0);
	
			case 'c':
							if(conf_filename!=NULL)
								free(conf_filename);
							conf_filename=strdup(optarg);
							break;
	
			case 'u':
							if(user_filename!=NULL)
								free(user_filename);
							user_filename=strdup(optarg);
							break;
			case 'U':
							if(user_uid!=NULL)
								free(user_uid);
							user_uid=strdup(optarg);
							break;
			case 'b':
							if(gl_bind_addr!=NULL)
								free(gl_bind_addr);
							if (is_ip(optarg))
								gl_bind_addr=strdup(optarg);
							else
							{
								gl_bind_addr=NULL;
								printf("Error invalid bind param\n");
							}
							break;
	
			case 'i':
							do_init=1;
							break;
	
			case 'm':
							add_missing=1;
							break;

			case 'd':
							if(perl_script_dir!=NULL)
								free(perl_script_dir);
							perl_script_dir=strdup(optarg);
							break;
	
			case 's':
							if(perl_script_main!=NULL)
								free(perl_script_main);
							perl_script_main=strdup(optarg);
							break;

#ifdef WITH_PYTHON
			case 'p':
							if ( python_script_dir != NULL )
								free( python_script_dir );
							python_script_dir = strdup( optarg );
							break;

			case 'x':
							if ( python_main_script != NULL )
								free( python_main_script );
							python_main_script = strdup( optarg );
							break;
#endif

			case 'n':
							new_hub_port=strtoul(optarg,NULL,10);
							if((new_hub_port<1)||(new_hub_port>65535))
							{
								fprintf(stderr,"invalid -n parameter\n");
								exit(1);
							}
							break;
			case 'f':
							tmp_reg_hub_port=strtoul(optarg,NULL,10);
							if((tmp_reg_hub_port<1)||(tmp_reg_hub_port>65535))
							{
								fprintf(stderr,"invalid -f parameter\n");
								exit(1);
							}
							reg_hub_port = tmp_reg_hub_port;
							break;


			case 'e':
							if(ext_prog_dir!=NULL)
								free(ext_prog_dir);
							ext_prog_dir=strdup(optarg);
							break;

			default:
							fprintf(stderr,"Unknown option: %c\n",ch);
							exit(1);
		}
	}

	/* check if required variables are filled ... and filled with correct values */

	/* ------------------------------- */
	/* 1) check the configuration file */
	/* ------------------------------- */
	if(conf_filename==NULL)
	{
		fprintf(stderr,"You must specify a configuration file.\n");
		exit(0);
	}

	if(stat(conf_filename,&st)==0) 
	{
		if(!S_ISREG(st.st_mode))
		{
			fprintf(stderr,"The configuration filename you gave is not a file.\n");
			exit(0);
		}
	}
	else
	{
		if((do_init|add_missing)==0)
		{
			fprintf(stderr,"The configuration filename you gave does not exist,\nyou must use --init to initialize it.\n");
			exit(0);
		}
	}

	{
		GString *tm;
		char *p2;

		tm=g_string_new(conf_filename);

		p2=strrchr(tm->str,'/');
		if(p2==NULL)
			tm=g_string_assign(tm,"tos_tbl");
		else
		{
			tm=g_string_truncate(tm,p2+1-tm->str);
			tm=g_string_append(tm,"tos_tbl");
		}
			
		tos_filename=tm->str;
		g_string_free(tm,FALSE);
	}

	/* ------------------------------- */
	/* 2) check the user file          */
	/* ------------------------------- */
	if(user_filename==NULL)
	{
		fprintf(stderr,"You must specify a user file.\n");
		exit(0);
	}

	if(stat(user_filename,&st)==0) 
	{
		if(!S_ISREG(st.st_mode))
		{
			fprintf(stderr,"The user filename you gave is not a file.\n");
			exit(0);
		}
	}
	else
	{
		if((do_init|add_missing)==0)
		{
			fprintf(stderr,"The user filename you gave does not exist,\nyou must use --init to initialize it.\n");
			exit(0);
		}
	}

	/* ------------------------------- */
	/* 3) check perl vars              */
	/* ------------------------------- */
	if( ((perl_script_dir!=NULL)&&(perl_script_main==NULL)) ||
		 ((perl_script_dir==NULL)&&(perl_script_main!=NULL)) )
	{
		fprintf(stderr,"To use perl script, you must use -d and -s at the same time (either both of them or none of them).\n");
		exit(0);
	}

	/* initialize file ? */
	if(do_init==1)
	{
		/* first: erase previous file */
		unlink(conf_filename);	
		unlink(user_filename);

		/* 2nd: create the default configuration file */
		conf_file=e_db_open(conf_filename);
		if(conf_file==NULL)
		{
			fprintf(stderr,"Unable to create %s\n",conf_filename);
			exit(0);
		}

		/* init the default variable */
		e_db_str_set(conf_file,"HUBNAME","noname");		/* hubname */
		e_db_str_set(conf_file,"HUBADDR","");				/* address of this server: (for DC Hublist) */
		e_db_str_set(conf_file,"HUBDESC","");				/* description of this server (for DC Hublist) */

		e_db_int_set(conf_file,"REG_HUB",0);				/* the server should be registered on DC Hublist server ? (0=No) */
		e_db_str_set(conf_file,"REG_SERV","vandel405.dynip.com");	/* address of the DC Hublist server */

		e_db_int_set(conf_file,"MAXUSER",100);				/* maximum number of users allowed on this hub */
		e_db_int_set(conf_file,"REDIR_FULL",0);			/* when the hub is full, redirect users to another hub ? (0=No) */
		e_db_str_set(conf_file,"REDIR_ADDR","");			/* address of the hub to redirect user to */

#if 0
		e_db_str_set(conf_file,"LOCK_PATTERN","^.*");	/* default denied nicknames: all nicknames */
#endif

		e_db_int_set(conf_file,"ONLY_REG",0);			/* this is a public hub */

		e_db_int_set(conf_file,"PUBCHAT",0);			/* public chat deactivated */

		e_db_int_set(conf_file,"MULTICHAT",1);			/* private chat deactivated */

		e_db_int_set(conf_file,"KICKTEMPBAN",0);		/* default ban time */

		e_db_int_set(conf_file,"MIN_SR_INTERVAL",10);		/* interval deetewn 2 search */

		e_db_int_set(conf_file,"SPAM_SILENT_LIMIT",10);		/* nbr message needed befor silence the user */

		e_db_int_set(conf_file,"SPAM_KICK_LIMIT",10);		/* nbr message needed befor diconnect the user */

		e_db_close(conf_file);
		conf_file=NULL;
		e_db_flush();

		/* 3rd: create the default user/passwd file */
		user_file=fopen(user_filename,"wb");
		if(user_file==NULL)
		{
			fprintf(stderr,"Unable to create %s\n",user_filename);
			exit(0);
		}

		/* user file format is quite simple: */
		/* ABBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBCCCCCCCCCCCCCD */
		/* where A='F' if the entry is empty and 'U' if it is used */
		/* B...B is a 32 bytes C-string (always ended by a '\0') containing a nickname. */
		/*        => the longest nickname takes 31 chars */
		/* C...C is a 16 butes C-string (always ended by a '\0') containing a password. */
		/*        => the longest password takes 15 chars */
		/* D is a 1 byte value: D='O' if this nick is an operator, D='N' if this nick is a normal user */
		/*        D='M' if this nick is the master of the hub (==operator + special function) */
		/*        only the master of the hub can modify configuration file remotely */
		{
			char master_user[]={'U',
									  'M','A','S','T','E','R',0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
									  'M','A','S','T','E','R',0,0,0,0,0,0,0,0,0,0,
									  'M'};

			if(fwrite(master_user,1,sizeof(master_user),user_file)!=sizeof(master_user))
			{
				fclose(user_file);
				unlink(user_filename);
				fprintf(stderr,"Unable to write inside %s. Disk full ? \n",user_filename);
				exit(0);
			}
		}

		fclose(user_file);
		user_file=NULL;
		fprintf(stderr,"Init done. Check the configuration file if you want to modify\n"
							"it. NEVER edit user file. After, you can start the hub without\n"
							"the --init option.\n"
							"To edit the configuration file, use edb_ed\n");
		fprintf(stderr,"This is the default configuration:\n"
							"HUBNAME => 'noname'\n"
							"HUBADDR => ''\n"
							"HUBDESC => ''\n"
							"REG_HUB => 0\n"
							"REG_SERV => 'vandel405.dynip.com'\n"
							"MAXUSER => 100\n"
							"REDIR_FULL => 0\n"
							"REDIR_ADDR => ''\n"
							"ONLY_REG => '0'\n"
							"PUBCHAT => '1'\n"
							"MULTICHAT => '1'\n"
							"KICKTEMPBAN => '0'\n"
							"MIN_SR_INTERVAL => '10'\n"
							"SPAM_SILENT_LIMIT => '10'\n"
							"SPAM_KICK_LIMIT => '10'\n"
							"LOCK_PATTERN -> '^#*' // remplace | by ! in this string\n"
							"\n"
							"In the user file, there is only 1 user who is the master of\n"
							"the hub. The nickname is 'MASTER' and its password is 'MASTER'\n"
							"\n");
		fprintf(stderr,"The easiest way to configure the hub is to start it and enter\n"
							"inside with the login/password 'MASTER'. Then, open a private\n"
							"chat with \"Hub-Security\" and use some commands of manual\n"
							"(see Documentation/commands.txt)\n"
							"\n");
		exit(0);
	}

	/* we are ready to start the hub */

	/* open the configuration file */
	conf_file=e_db_open(conf_filename);
	if(conf_file==NULL)
	{
		fprintf(stderr,"Unable to open %s\n",conf_filename);
		exit(0);
	}

	if(new_hub_port!=-1)
	{
		/* if a port is provided, save the port number into the database */
		printf ("update HUBPORT to %d\n", new_hub_port);
		e_db_int_set(conf_file,"HUBPORT",new_hub_port);
	}

	/* initialize hub_port param to the saved value or to the default 411 value */
	if(!e_db_int_get(conf_file,"HUBPORT",&new_hub_port))
		new_hub_port=411;
	hub_port=new_hub_port;

	/* open the user file */
	user_file=fopen(user_filename,"r+b");
	if(user_file==NULL)
	{
		fprintf(stderr,"Unable to open %s\n",user_filename);
		exit(0);
	}

	/* automatically starts some external programs */
	if(ext_prog_dir!=NULL)
	{
		GString *str;
		FILE *f;

		str=g_string_new("");
		g_string_sprintf(str,"%s/AUTOSTART",ext_prog_dir);
		f=fopen(str->str,"rt");
		if(f!=NULL)
		{
			char buf[2048];

			while(fgets(buf,sizeof(buf),f)!=NULL)
			{
				char *r;

				r=strpbrk(buf,"\r\n");
				if(r!=NULL)
					*r='\0';

				add_prog_to_start(buf);
			}
			
			fclose(f);
		}

		g_string_free(str,TRUE);
	}
}

/* #define DEBUG_MEMORY */

#ifdef DEBUG_MEMORY
#include <mcheck.h>
static void
enable (int sig)
{
	mtrace ();
	signal (SIGUSR1, enable);
}

static void
disable (int sig)
{
	muntrace ();
	signal (SIGUSR2, disable);
}
#endif /* DEBUG_MEMORY */


int	cnxs_fnc_process_io(fd_set *rd, fd_set *wd, CNX_ENTRY *ptr);
int	cnxs_fnc_process_cmd(CNX_ENTRY *ptr);
int	cnxs_fnc_process_mask(CNX_ENTRY *ptr, int n, fd_set *rd, fd_set *wd);

int load_dyn_conf(void)
{
	char	*hub_name;
	char	*pattern_list;

	G_LOCK(conf_file);
	hub_name = e_db_str_get(conf_file,"HUBNAME");
	if (!e_db_int_get(conf_file,"MIN_SR_INTERVAL" , &gl_sr_interval))
	{
		gl_sr_interval = 10;
	}
	if (!e_db_int_get(conf_file,"SPAM_SILENT_LIMIT" , &gl_silent_limit))
	{
		gl_silent_limit = 10;
	}
	if (!e_db_int_get(conf_file,"SPAM_KICK_LIMIT" , &gl_kick_limit))
	{
		gl_kick_limit = 10;
	}
	if ((pattern_list = e_db_str_get(conf_file,"LOCK_PATTERN")) != NULL)
	{
		char	*tmp;
			
		if (gl_lock_patern)
		{
			regfree(gl_lock_patern);
			free(gl_lock_patern);
			gl_lock_patern = NULL;
		}

		if (*pattern_list != '\0')
		{
			while ((tmp = strchr(pattern_list, '')) != NULL)
				*tmp = '|';
			
			gl_lock_patern = malloc (sizeof(regex_t));
			if (regcomp(gl_lock_patern, pattern_list, REG_NOSUB))
			{
				free(gl_lock_patern);
				gl_lock_patern = NULL;
			}
		}
		free(pattern_list);
	}
	else
	{
		if (gl_lock_patern)
		{
			regfree(gl_lock_patern);
			free(gl_lock_patern);
			gl_lock_patern = 0;
		}
	}
	G_UNLOCK(conf_file);

	if (hubname != NULL)
		g_string_free(hubname, TRUE);	
	hubname=g_string_new("$HubName ");

	if (hub_name != NULL)
	{
		hubname=g_string_append(hubname,hub_name);
		free(hub_name);
	}
	hubname=g_string_append(hubname,"|");
	return (0);
}

int main(int argc,char **argv)
{
	int omega1=0,omega2=0,omega3=0;
	time_t last_time;
	int i;

#ifdef DEBUG_MEMORY
	mcheck(0);
#endif /* DEBUG_MEMORY */

	printf("DChub v" VERSION "\n");
	parse_command_line_options(argc,argv);

#ifndef __CYGWIN__
	g_thread_init(NULL);
#endif

	cnx_lst=new_cnx_hash_table();
	route_table = new_route_table();
	/* ignore some signals */
	set_sig();
#ifdef DEBUG_MEMORY
	signal (SIGUSR1, enable);
	signal (SIGUSR2, disable);
#endif /* DEBUG_MEMORY */

	load_dyn_conf();
	load_tos_data(tos_filename);

	/* create the hub socket */
	main_sck=_x_tcp(hub_port);
	if(main_sck==-1)
	{
		fprintf(stderr,"Unable to create hub port.\n");
		exit(1);
	}
	listen(main_sck,64);

	if(init_multihub_side())
	{
		fprintf(stderr,"Unable to initialize multihub side.\n");
		exit(1);
	}
	if (user_uid)
	{
		struct passwd* pw;
		struct passwd* new_pw;

		new_pw = getpwnam(user_uid);
		if (new_pw)
		{
			if (setuid(new_pw->pw_uid))
				perror("setuid");
		}
		pw = getpwuid(getuid());
		if (pw)
			printf("dchub is now running as %s\n", pw->pw_name);
	}
	init_perl();
#ifdef WITH_PYTHON
	init_python();
#endif
#if 0
	{
		int i;
		for(i=0;i<10000;i++)
		{
			printf("%d\n",i);
			send_evt_to_script("test","the emitter",3,"argument1","argument2","argument3");
#ifdef WITH_PYTHON
			send_evt_to_pscript("test","the emitter",3,"argument1","argument2","argument3");
#endif
		}
	}

	exit(0);
#endif

	/* begin of the main loop */
	while(!user_wants_to_quit)
	{
		fd_set rd,wd;
		struct timeval tv;
		int ln;
		int n=-1;

		if ((omega1 != cnx_lst->wait_len) || (omega2 != cnx_lst->len) || (omega3 != route_table->server_count))
		  /*  printf("there is " COLOMAJ1 "%d" COLONONE " user in queue, " COLOMAJ1 "%d" COLONONE" user login, "COLOMAJ1 "%d" COLONONE" hub in cluster\n", cnx_lst->wait_len, cnx_lst->len, route_table->server_count); */
		  ;
		omega1 = cnx_lst->wait_len;
		omega2 = cnx_lst->len;
		omega3 = route_table->server_count;

		FD_ZERO(&rd);
		FD_ZERO(&wd);
	
		FD_SET(main_sck,&rd);		/* we want someone to come in */
		n=max(n,main_sck);

		/********************************************/
		/**********   CREATE MASK  ******************/
		/********************************************/

		/* find max id socket*/
		/* for loged users */
		for(i=0;i<HASH_SIZE;i++)
		{
			CNX_ENTRY *ptr;
			for (ptr=cnx_lst->entry[i]; ptr; ptr=ptr->next)
				if (ptr->sock_fd != -1)
			n = cnxs_fnc_process_mask(ptr, n, &rd, &wd);
		}

		/* for queued users */
		{
			CNX_ENTRY *ptr;
			for (ptr=cnx_lst->wait; ptr; ptr=ptr->next)
				n = cnxs_fnc_process_mask(ptr, n, &rd, &wd);
		}

		/* Cluster Part 6 */
		for (i=1;i<=route_table->max_server_id;i++)
		{
			if (route_table->table[i].connected)
			{
				n = cnxs_fnc_process_mask(route_table->table + i, n,&rd, &wd);
			}
		}

		/********************************************/
		/********************************************/

		/* wait at most 1 second */
		tv.tv_sec=1;
		tv.tv_usec=0;

		ln=select(n+5,&rd,&wd,NULL,&tv);
		
		if (ln == -1)
		{
			printf(COLOYEL1"error in file desc, user_wants_to_quit = %d"COLONONE"\n", user_wants_to_quit);
			user_wants_to_quit = 1;
		}
		if(ln>0)
		{
			if(FD_ISSET(main_sck,&rd))
				new_incoming_connexion();
			/********************************************/
			/*********** IO PROCESS *********************/
			/********************************************/

			/* incomming */
			/* for loged users */
			HL_LOCK_WRITE(cnx_struct_lock);

			one_more_time:
			for(i=0;i<HASH_SIZE;i++)
			{
				CNX_ENTRY *ptr;
				for(ptr=cnx_lst->entry[i]; ptr; ptr=ptr->next)
				if (ptr->sock_fd != -1)
				{
					if (cnxs_fnc_process_io(&rd, &wd, ptr)) /* if the entry had been remove resume form the begining */
						goto one_more_time;
				}
			}
		
			/* for queued users */
			one_more_time2:
			{
				CNX_ENTRY *ptr;
				for(ptr=cnx_lst->wait; ptr; ptr=ptr->next)
				{
					if (cnxs_fnc_process_io(&rd, &wd, ptr)) /* if the entry had been remove resume form the begining */
						goto one_more_time2;
				}
			}
			HL_UNLOCK_WRITE(cnx_struct_lock);
	
			one_more_time3: 
			for (i=1;i<=route_table->max_server_id;i++)/* for others servers */
			{
				if (route_table->table[i].connected)
				{
					if (cnxs_fnc_process_io(&rd, &wd, (route_table->table + i)))
						goto one_more_time3;
				}
			}
		} /* end ln > 0 */
		
		/* time to handle all queued command */
		/* for loged users */

			/********************************************/
			/********** CMD PROCESS *********************/
			/********************************************/

		HL_LOCK_WRITE(cnx_struct_lock);
	again:
		for(i=0;i<HASH_SIZE;i++)
		{
			CNX_ENTRY *ptr;
			for(ptr=cnx_lst->entry[i]; ptr; ptr=ptr->next)
			{
				if (ptr->sock_fd != -1)
					if (cnxs_fnc_process_cmd(ptr))
						goto again;
			}
		}
		
		/* for queued users */
	again2:
		{
			CNX_ENTRY *ptr;
			
			for (ptr=cnx_lst->wait; ptr; ptr=ptr->next)
			{
				if (new_cnx(ptr)==0)
					goto again2;
			}
		}
		HL_UNLOCK_WRITE(cnx_struct_lock);
		

		/* Cluster Part 8 */
		for (i=1;i<=route_table->max_server_id;i++)
		{
			if (route_table->table[i].connected)
			{
				if (cnxs_fnc_process_cmd(route_table->table + i))
					goto again;
			}
		}
		
		/***************************************************/
		/************* TIME OUT PROCESS ********************/
		/***************************************************/

		/* abort connection taking too much time to start */ 
		/* for queued users only */
		last_time = gl_cur_time;
		gl_cur_time=time(NULL);
		if (last_time != gl_cur_time) /* do not revefie the timeout more than 2 time in the same second */
		{
			CNX_ENTRY *ptr;
			
			for(ptr=cnx_lst->wait; ptr; ptr=ptr->next)
			{
				if((gl_cur_time-ptr->start_time)>10)	/* wait up to 10 seconds to perform handshake */
					shutdown(ptr->sock_fd,2);
			}
		}
#ifndef __CYGWIN__
		check_hub_registration();
#endif
		check_prog_to_start();
		timeout_tos();
	} /* end of the main loop */
	printf("ok\n");
	
	/***************************************************/
	/************** END close soft ********************/
	/***************************************************/


	if(conf_file!=NULL)
	{
		G_LOCK(conf_file);
		e_db_close(conf_file);
		conf_file=NULL;
		e_db_flush();
	}

	if(user_file!=NULL)
	{
		G_LOCK(user_file);
		fclose(user_file);
		user_file=NULL;
	}

	save_tos_data(tos_filename);
	exit_perl();
#ifdef WITH_PYTHON
	exit_python_interpreter();
#endif
	exit(0);
}
